@@ -1,11 +1,12 @@
 from concurrent import futures
-import random, json, os, tempfile, shutil
+import random, json, os, tempfile, shutil, importlib.util
 
 from scipy import ndimage
 import numpy as np
 import tensorflow as tf
 from scipy import misc
 import openpyxl
+import imgaug as ia
 
 class Dataset(object):
     """A dataset consisting of some examples of input/output pairs. Minimizes
@@ -15,8 +16,6 @@ class Dataset(object):
         then these examples will be ignored.
     """
 
-    SEGMENT_POOL_MULTIPLIER = 5
-
     def __init__(self, path, segment_size=None):
         """Create a dataset.
 
@@ -39,6 +38,34 @@ class Dataset(object):
             self._segments = 0
             self._save_metadata()
 
+    def add(self, src):
+        """Add the examples from a source dataset to this dataset.
+
+        Args:
+            src (Dataset): The source dataset.
+        """
+
+        def gen():
+            for segment in range(src._segments):
+                inputs, outputs = src._load_segment(segment)
+                for i in range(inputs.shape[0]):
+                    yield inputs[i, ...], outputs[i, ...]
+
+        self._load_examples(gen())
+
+    def load(self, loader_path):
+        """Load examples using a loader.
+
+        Args:
+            loader_path (str): A path to a file that defines a function named
+                ``load``. This function must yield the examples of the
+                dataset.
+        """
+        spec = importlib.util.spec_from_file_location("loader", loader_path)
+        module = importlib.util.module_from_spec(spec)
+        spec.loader.exec_module(module)
+        self._load_examples(module.load())
+
     def initialize_from_aspects(self, path, transform):
         """Initialize by reading some examples' aspects and generating inputs
             and outputs from them.
@@ -62,18 +89,19 @@ class Dataset(object):
             if not dir_ent.is_dir():
                 return
             aspects = {}
-            with os.scandir(dir_ent.path) as example_dir_it:
-                for example_dir_ent in example_dir_it:
-                    if example_dir_ent.is_dir():
-                        continue
-                    name, _ = os.path.splitext(example_dir_ent.name)
-                    name = name.lower()
-                    val = self._read_aspect(example_dir_ent.path)
-                    aspects[name] = val
+            for example_dir_ent in os.scandir(dir_ent.path):
+                if example_dir_ent.is_dir():
+                    continue
+                name, _ = os.path.splitext(example_dir_ent.name)
+                name = name.lower()
+                val = self._read_aspect(example_dir_ent.path)
+                aspects[name] = val
             return transform(aspects)
         inputs, outputs = [], []
         with futures.ThreadPoolExecutor() as executor:
             for example in executor.map(get_example, os.scandir(path)):
+                if example is None:
+                    continue
                 inp, out = example
                 inputs.append(inp)
                 outputs.append(out)
@@ -262,41 +290,64 @@ class Dataset(object):
         """
         return self._segments * self._segment_size
 
-    def split(self, p, path_larger, path_smaller):
+    def split(self, p, path_larger, path_smaller, seed=None):
         """Split this dataset.
 
-        Note: Will assign entire segments to one part of the split or the other.
-            Result may not approximate p well if there aren't many segments.
-            If the number of segments is very small, there is the possibility
-            for errors.
+        The larger dataset will be given a segment size that is ``1 - p`` times
+        this dataset's segment size, and likewise with ``p`` for the smaller
+        dataset. Chooses the examples to allocate to each side of the split
+        at random.
 
         Args:
-            p (float): The proportion of the dataset to put in the smaller part
+            p (float): The proportion of the examples to put in the smaller part
                 of the split.
             path_larger (str): The path to a directory in which to store the
                 larger part of the split.
             path_smaller (str): The path to a directory in which to store the
                 smaller part of the split.
+            seed (int): A number to seed the random number generator with before
+                randomly assigning examples to splits. With the same ``seed``,
+                two ``split`` operations should have the same effect.
 
         Returns:
-            (tuple of dataset.Dataset): The two datasets that result from the
-                split. The smaller one is last.
+            tuple(dataset.Dataset, dataset.Dataset): The two datasets that
+            result from the split. The smaller one is last.
         """
-        smaller = Dataset(path_smaller, int(self._segment_size))
-        larger = Dataset(path_larger, int(self._segment_size))
-        smaller_limit = int(p * self._segments)
-        for i in range(smaller_limit):
-            src = self._get_segment_path(i)
-            dst = smaller._get_segment_path(i)
-            shutil.copytree(src, dst)
-        smaller._segments = smaller_limit
-        smaller._save_metadata()
-        for i in range(smaller_limit, self._segments):
-            src = self._get_segment_path(i)
-            dst = larger._get_segment_path(i - smaller_limit)
-            shutil.copytree(src, dst)
-        larger._segments = self._segments - smaller_limit
-        larger._save_metadata()
+        def new_segment_size(p):
+            return max(1, round(self._segment_size * p))
+
+        larger = Dataset(path_larger, new_segment_size(1 - p))
+        smaller = Dataset(path_smaller, new_segment_size(p))
+
+        def example_generator(include_example):
+            i = 0
+            for segment in range(self._segments):
+                inputs_path = self._get_segment_file_path(segment, "inputs.npy")
+                outputs_path = self._get_segment_file_path(segment,
+                                                           "outputs.npy")
+                inputs = np.load(inputs_path)
+                outputs = np.load(outputs_path)
+                for segment_i in range(inputs.shape[0]):
+                    if include_example(i):
+                        yield (inputs[segment_i, ...], outputs[segment_i, ...])
+                    i += 1
+
+        num_examples = self._segments * self._segment_size
+        num_examples_smaller = round(num_examples * p)
+        if seed is not None:
+            random.seed(seed)
+        chosen_for_smaller = set(random.sample(range(num_examples),
+                                               num_examples_smaller))
+
+        def include_example_larger(i):
+            return i not in chosen_for_smaller
+
+        larger._load_examples(example_generator(include_example_larger))
+
+        def include_example_smaller(i):
+            return i in chosen_for_smaller
+
+        smaller._load_examples(example_generator(include_example_smaller))
         return larger, smaller
 
     def map(self, fn):
@@ -419,19 +470,28 @@ class Dataset(object):
         self._segments = scratch_dataset._segments
         self._save_metadata()
 
-    def get_batch(self, size):
-        """Get a batch of examples.
+    def get_batch(self, size, pool_multiplier=5):
+        """Get a batch of examples, randomly selected from a pool.
 
         Args:
             size (int): The number of examples.
+            pool_multiplier (int): If ``n`` segments are needed to get ``size``
+                examples, then the examples will be drawn at random from
+                ``pool_multiplier * n`` segments. This result is then capped
+                at the number of segments that exist. The default is ``5``.
 
         Returns:
-            (tuple of np.ndarray): The inputs and outputs of the batch. Each
-                array has size rows, where the i-th row corresponds to the i-th
-                example.
+            tuple(numpy.ndarray, numpy.ndarray): The inputs and outputs of
+            the batch. Each array has ``size`` rows, where the ``i``-th row
+            corresponds to the ``i``-th example.
         """
+        assert isinstance(pool_multiplier, int), "argument pool_multiplier " \
+                                                 "must be an int"
+        assert pool_multiplier >= 1, "argument pool_multiplier must be at " \
+                                     "least 1"
+
         segments_needed = (int(size / self._segment_size) + 1)
-        segments_needed *= self.SEGMENT_POOL_MULTIPLIER
+        segments_needed *= pool_multiplier
         segments_needed = min(segments_needed, self._segments)
         chosen_segments = random.sample(range(self._segments), segments_needed)
         cache = []
@@ -454,6 +514,121 @@ class Dataset(object):
         outputs = np.stack(outputs, axis=0)
         return inputs, outputs
 
+    class BatchIterator:
+        def __init__(self, dataset, batch_size, pool_size):
+            self._dataset = dataset
+            self._batch_size = batch_size
+            self._pool_size = pool_size
+
+        def __next__(self):
+            return self._dataset.get_batch(self._batch_size, self._pool_size)
+
+        def __iter__(self):
+            return self
+
+    class EpochBatchIterator:
+        def __init__(self, dataset, batch_size, pool_size):
+            self._dataset = dataset
+            self._batch_size = batch_size
+            self._pool_size = pool_size
+
+            self._epoch = 0
+            self._start_epoch()
+            inputs, outputs = dataset.get_batch(2, 1)
+            segments = (int(batch_size / dataset._segment_size) + 1)
+            segments *= pool_size
+            segments = min(segments, dataset._segments)
+            self._pool_max = (segments + 2) * dataset._segment_size
+            self._pool_min = segments * dataset._segment_size
+            self._pool_inputs = np.empty((self._pool_max,) + inputs.shape[1:])
+            self._pool_outputs = np.empty((self._pool_max,) + outputs.shape[1:])
+
+        def __next__(self):
+            self._replenish()
+            chosen = np.random.choice(self._pool_top + 1, self._batch_size)
+            inputs = self._pool_inputs[chosen, ...]
+            outputs = self._pool_outputs[chosen, ...]
+            new_inputs = np.delete(self._pool_inputs[:self._pool_top, ...],
+                                   chosen, axis=0)
+            new_outputs = np.delete(self._pool_outputs[:self._pool_top, ...],
+                                    chosen, axis=0)
+            self._pool_top = new_inputs.shape[0]
+            self._pool_inputs[:self._pool_top, ...] = new_inputs
+            self._pool_outputs[:self._pool_top, ...] = new_outputs
+            return inputs, outputs
+
+        def __iter__(self):
+            return self
+
+        def _replenish(self):
+            while (self._pool_top + 1) <= self._pool_min:
+                try:
+                    segment = self._segment_use_map.index(False)
+                except ValueError:
+                    break
+                self._segment_use_map[segment] = True
+                inputs, outputs = self._dataset._load_segment(segment)
+                new_pool_top = self._pool_top + inputs.shape[0]
+                self._pool_inputs[self._pool_top:new_pool_top, ...] = inputs
+                self._pool_outputs[self._pool_top:new_pool_top, ...] = outputs
+                self._pool_top = new_pool_top
+            if (self._pool_top + 1) < self._batch_size:
+                self._start_epoch()
+                self._replenish()
+
+        def _start_epoch(self):
+            self._epoch += 1
+            self._pool_top = 0
+            self._segment_use_map = [False for _
+                                     in range(self._dataset._segments)]
+
+        def get_epoch(self):
+            return self._epoch
+
+    def get_batch_iterable(self, batch_size, pool_size=5, epochs=False):
+        """Get an iterable over batches of this dataset's examples.
+
+        Args:
+            batch_size (int): The number of examples per batch.
+            pool_size (int): The minimum number of segments to pool together in
+                memory to draw random batches from. If omitted, ``5``.
+            epochs (bool): Whether to organize the batches into epochs, so that
+                every example is seen once before any is  repeated,
+                then every example is seen twice before any is repeated a third
+                time, etc. If omitted, ``False``.
+
+        Returns:
+            (Iterable): An iterable over batches of this dataset's examples.
+                Each element of the sequence is a ``tuple`` of
+                ``numpy.ndarray``s, where the first array is the inputs and the
+                second array is the outputs.
+        """
+        if not epochs:
+            return Dataset.BatchIterator(self, batch_size, pool_size)
+        else:
+            return Dataset.EpochBatchIterator(self, batch_size, pool_size)
+
+    def get_all(self):
+        """Get all examples in the dataset.
+
+        Returns the examples in the same order each time it is called.
+
+        Returns:
+            (tuple(numpy.ndarray, numpy.ndarray)): A tuple of two
+                elements: the inputs of all the examples and the outputs of
+                all the examples. These will have shape (num_examples, ...).
+        """
+        inputs, outputs = [], []
+        for i in range(self._segments):
+            inp, out = self._load_segment(i)
+            inputs.append(inp)
+            outputs.append(out)
+        inputs = np.concatenate(inputs, axis=0)
+        outputs = np.concatenate(outputs, axis=0)
+        assert inputs.shape[0] == outputs.shape[0], "Something is wrong!" \
+            "There are not equal numbers of inputs and outputs in this dataset."
+        return (inputs, outputs)
+
     def get_data_fn(self, batch_size, num_batches):
         """Get a data function for this dataset.
 
@@ -472,14 +647,47 @@ class Dataset(object):
             batch_size, num_batches, shuffle=False,
             queue_capacity=num_batches)
 
-    def close(self):
-        """Close this dataset. This dataset will not be useable afterward.
+    def delete(self):
+        """Delete this dataset.
 
-        This method must be called because it deletes temporary files on disk.
-        """
+        Deletes this dataset's directory on disk. This will cause any files
+        in that directory, even those not created by this dataset,
+        to be deleted. This dataset will not be usable afterward."""
         shutil.rmtree(self._path)
         self._path = None
 
+    def augment(self, augmenter=None, input_augmenter=None,
+                output_augmenter=None):
+        """Augment this dataset.
+
+        Applies transformations to the examples in this dataset to produce
+            additional examples.
+
+        Args:
+            augmenter (imgaug.augmenters.Augmenter): An augmenter.
+            augment_inputs (bool): Whether to apply the augmenter to the inputs.
+            augment_outputs (bool): Whether to apply the augmenter to the
+                outputs.
+        """
+        original_segments = self._segments
+        self._segments = 0
+        for i in range(original_segments):
+            inputs, outputs = self._load_segment(i)
+            if augmenter is not None:
+                det_augmenter = augmenter.to_deterministic()
+                gen = det_augmenter.augment_batches([inputs])
+                inputs = list(gen)[0]
+                gen = det_augmenter.augment_batches([outputs])
+                outputs = list(gen)[0]
+            if input_augmenter is not None:
+                gen = input_augmenter.augment_batches([inputs])
+                inputs = list(gen)[0]
+            if output_augmenter is not None:
+                gen = output_augmenter.augment_batches([outputs])
+                outputs = list(gen)[0]
+            shutil.rmtree(self._get_segment_path(self._segments))
+            self._add_segment(inputs, outputs)
+
     def _load_examples(self, example_iterator):
         inputs, outputs = [], []
         for i, example in enumerate(example_iterator):
